# coding=utf-8
import atexit
import os
import re
import select
import time
from subprocess import Popen, PIPE


TENSORBOARD_BINARY = "tensorboard"


class Process:
    instances = []  # type: List[Process]

    def __init__(self, command):
        if type(command == list):
            self.command = command
        else:
            self.command = command.split(" ")

        self.proc = None  # type: Popen

    def run(self):
        environment = os.environ.copy()
        # necessary for reading from processes that don't flush
        # stdout automatically
        environment["PYTHONUNBUFFERED"] = "1"
        self.proc = Popen(self.command, env=environment, stdout=PIPE)
        Process.instances.append(self)

    def is_running(self):
        return self.proc is not None and not bool(self.proc.poll())

    def read_line(self, time_limit=None):
        if self.proc is not None:
            poll_obj = select.poll()
            poll_obj.register(self.proc.stdout, select.POLLIN)
            start = time.time()
            while time_limit is None or time.time() - start < time_limit:
                poll_result = poll_obj.poll(0)
                if poll_result:
                    line = self.proc.stdout.readline().decode()
                    return line
            raise TimeoutError()
        else:
            return None

    def terminate(self, wait=False):
        if self.proc is not None:
            self.proc.stdout.close()
            self.proc.terminate()
            if wait:
                self.proc.wait()

    def pid(self):
        if self.proc is not None:
            return self.proc.pid
        else:
            return None

    @staticmethod
    def terminate_all(wait=False):
        """

        :param wait: Wait for each to terminate
        :type wait: bool
        :return:
        :rtype:
        """
        for instance in Process.instances:
            if instance.is_running():
                instance.terminate(wait)

    @staticmethod
    def create_process(command):
        if getattr(select, "poll", None) is not None:
            return Process(command)
        else:
            return WindowsProcess(command)


class WindowsProcess(Process):
    def __init__(self, command):
        Process.__init__(self, command)

    def read_line(self, time_limit=None):
        """ Time limit has no effect.
         The operation will always block on Windows."""
        if self.proc is not None:
            return self.proc.stdout.readline().decode()
        else:
            return None


# Kill all on exit
atexit.register(Process.terminate_all)


class ProcessError(Exception):
    pass


class TensorboardNotFoundError(ProcessError):
    pass


class UnexpectedOutputError(ProcessError):
    def __init__(self, output, expected=None):
        self.expected = expected
        self.output = output


def stop_all_tensorboards():
    for process in Process.instances:
        print("Process '%s', running %d" % (process.command[0],
                                            process.is_running()))
        if process.is_running() and process.command[0] == "tensorboard":
            process.terminate()


def run_tensorboard(logdir, listen_on="0.0.0.0", tensorboard_args=None):
    if tensorboard_args is None:
        tensorboard_args = []
    tensorboard_instance = Process.create_process(
        TENSORBOARD_BINARY.split(" ")
        + ["--logdir", logdir, "--host", listen_on] + tensorboard_args)
    try:
        tensorboard_instance.run()
    except FileNotFoundError as ex:
        raise TensorboardNotFoundError(ex)

    # Read first line of output from tensorboard - it should contain
    # the port where it listens
    data = tensorboard_instance.read_line(time_limit=10)
    search = re.search("on port ([0-9]+)", data)
    if search is not None:
        port = search.group(1)
        return port
    else:
        tensorboard_instance.terminate()
        raise UnexpectedOutputError(
            data,
            expected="The port Tensorboard that listens on."
            )



# coding=utf-8
import re
from pathlib import Path

from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request, Response, redirect, url_for

import sacredboard.app.process.process as proc
from ..process import process

routes = Blueprint("routes", __name__)


@routes.route("/")
def index():
    return redirect(url_for("routes.show_runs"))


@routes.route("/runs")
def show_runs():
    # return render_template("runs.html", runs=data.runs(), type=type)
    return render_template("runs.html", runs=[], type=type)


@routes.route("/api/run")
def api_runs():
    data = current_app.config["data"]
    # TODO: Move the logic somewhere else?
    draw = request.args.get("draw")
    draw = int(draw) if draw is not None else 1
    start = 0 if not request.args.get("start") \
        else int(request.args.get("start"))
    length = -1 if not request.args.get("length") \
        else int(request.args.get("length"))
    length = length if length >= 0 else None

    # TODO: Make it work with heartbeat_diff
    order_column = request.args.get("order[0][column]")
    order_dir = request.args.get("order[0][dir]")
    if order_column is not None:
        order_column = \
            request.args.get("columns[%d][name]" % int(order_column))
        if order_column == "hostname":
            order_column = "host.hostname"
    runs = data.get_runs(start=start, limit=length,
                         sort_by=order_column, sort_direction=order_dir)
    # records_total should be the total size of the records in the database,
    # not what was returned
    records_total = runs.count()
    records_filtered = runs.count()
    return Response(render_template("api/runs.js", runs=runs,
                                    draw=draw, recordsTotal=records_total,
                                    recordsFiltered=records_filtered),
                    mimetype="application/json")


@routes.route("/api/run/<run_id>")
def api_run(run_id):
    data = current_app.config["data"]
    run = data.get_run(run_id)
    records_total = 1 if run is not None else 0
    records_filtered = records_total
    return Response(render_template("api/runs.js", runs=[run], draw=1,
                                    recordsTotal=records_total,
                                    recordsFiltered=records_filtered,
                                    full_object=True),
                    mimetype="application/json")


@routes.route("/tensorboard/start/<run_id>/<int:tflog_id>")
def run_tensorboard(run_id, tflog_id):
    data = current_app.config["data"]
    # optimisticaly suppose the run exists...
    run = data.get_run(run_id)
    base_dir = Path(run["experiment"]["base_dir"])
    log_dir = Path(run["info"]["tensorflow"]["logdirs"][tflog_id])
    # TODO ugly!!!
    if log_dir.is_absolute():
        path_to_log_dir = log_dir
    else:
        path_to_log_dir = base_dir.joinpath(log_dir)

    port = int(proc.run_tensorboard(str(path_to_log_dir)))
    url_root = request.url_root
    url_parts = re.search("://([^:/]+)", url_root)
    redirect_to_address = url_parts.group(1)
    return redirect("http://%s:%d" % (redirect_to_address, port))


@routes.route("/tensorboard/stop")
def close_tensorboards():
    proc.stop_all_tensorboards()
    return "Stopping tensorboard"


@routes.errorhandler(process.TensorboardNotFoundError)
def handle_tensorboard_not_found(e):
    return "Tensorboard not found in your system." \
           " Please install tensorflow first. Sorry.", 503


@routes.errorhandler(TimeoutError)
def handle_tensorboard_timeout(e):
    return "Tensorboard does not respond. Sorry.", 503


@routes.errorhandler(process.UnexpectedOutputError)
def handle_tensorboard_unexpected_output(e: process.UnexpectedOutputError):
    return "Tensorboard outputted '%s'," \
           " but the information expected was: '%s'. Sorry."\
           % (e.output, e.expected), 503


def setup_routes(app):
    app.register_blueprint(routes)



# coding=utf-8
import locale

import click
from flask import Flask
from gevent.pywsgi import WSGIServer

from sacredboard.app.config import jinja_filters
from sacredboard.app.config import routes
from sacredboard.app.data.mongodb import PyMongoDataAccess

locale.setlocale(locale.LC_ALL, '')
app = Flask(__name__)


@click.command()
@click.option("-m", default="sacred", metavar="CONNECTION_STRING",
              help="Connect to MongoDB on host:port:database or database. "
                   "Default: sacred")
@click.option("--no-browser", is_flag=True, default=False,
              help="Do not open web browser automatically.")
@click.option("--debug", is_flag=True, default=False,
              help="Run the application in Flask debug mode "
                   "(for development).")
@click.version_option()
def run(debug, no_browser, m):
    """
    \b
    Sacredboard is a monitoring dashboard for Sacred.
    Homepage: http://github.com/chovanecm/sacredboard

    Example usage:

        \b
        sacredboard -m sacred
            Starts Sacredboard on default port (5000) and connects to
            a local MongoDB database called 'sacred'. Opens web browser.
            Note: MongoDB must be listening on localhost.
        \b
        sacredboard -m 192.168.1.1:27017:sacred
            Starts Sacredboard on default port (5000) and connects to
            a MongoDB database running on 192.168.1.1 on port 27017
            to a database called 'sacred'. Opens web browser.

    """
    add_mongo_config(app, m)
    app.config['DEBUG'] = debug
    app.debug = debug
    jinja_filters.setup_filters(app)
    routes.setup_routes(app)
    app.config["data"].connect()
    if debug:
        app.run(host="0.0.0.0", debug=True)
    else:
        for port in range(5000, 5050):
            http_server = WSGIServer(('0.0.0.0', port), app)
            try:
                http_server.start()
            except OSError as e:
                    # try next port
                    continue
            print("Starting sacredboard on port %d" % port)
            if not no_browser:
                click.launch("http://127.0.0.1:%d" % port)
            http_server.serve_forever()
            break


def add_mongo_config(app, connection_string):
    """
    Configure the app to use MongoDB.

    :param app: the Flask Application
    :type app: Flask
    :param connection_string: in format host:port:database or database
            (default: sacred)
    :type connection_string: str
    """
    split_string = connection_string.split(":")
    config = {"host": "localhost", "port": 27017, "db": "sacred"}
    if len(split_string) > 0 and len(split_string[-1]) > 0:
        config["db"] = split_string[-1]
    if len(split_string) > 1:
        config["port"] = int(split_string[-2])
    if len(split_string) > 2:
        config["host"] = split_string[-3]
    app.config["data"] = PyMongoDataAccess(
        config["host"], config["port"], config["db"])


if __name__ == '__main__':
    run()



# coding=utf-8
import bson
import pymongo


class PyMongoDataAccess:
    """ Access records in MongoDB. """
    def __init__(self, host, port, db):
        self._host = host
        self._port = port
        self._db_name = db
        self._client = None
        self._db = None

    def connect(self):
        self._client = pymongo.MongoClient(host=self._host, port=self._port)
        self._db = getattr(self._client, self._db_name)

    def get_runs(self, sort_by=None, sort_direction=None, start=0, limit=None):
        cursor = self._db.runs.find()
        if sort_by is not None:
            cursor = self._apply_sort(cursor, sort_by, sort_direction)
        cursor = cursor.skip(start)
        if limit is not None:
            cursor = cursor.limit(limit)
        return cursor

    def get_run(self, run_id):
        try:
            cursor = self._db.runs.find({"_id": int(run_id)})
        except ValueError:
            # Probably not a number.
            cursor = self._db.runs.find({"_id": bson.ObjectId(run_id)})
        run = None
        for c in cursor:
            run = c
        return run

    def _apply_sort(self, cursor, sort_by, sort_direction):
        """
        :type sort_direction: str
        :return:
        :rtype: pymongo.Cursor
        """
        if sort_direction is not None and sort_direction.lower() == "desc":
            sort = pymongo.DESCENDING
        else:
            sort = pymongo.ASCENDING
        return cursor.sort(sort_by, sort)
